import numpy as np
import warnings


class ReferenceImputer:
    '''
    Impute features using reference values.

    Args:
      reference: the reference value for replacing missing features.
    '''
    def __init__(self, reference):
        if reference.ndim == 1:
            reference = reference[np.newaxis]
        elif reference.ndim == 2:
            if reference.shape[0] > 1:
                raise ValueError('expecting one set of reference values, '
                                 'got {}'.format(reference.shape[0]))
        self.reference = reference
        self.reference_repeat = self.reference
        self.samples = 1
        self.num_groups = reference.shape[-1]

    def __call__(self, x, S):
        # Repeat reference.
        if len(self.reference_repeat) != len(x):
            self.reference_repeat = self.reference.repeat(len(x), 0)

        x_ = x.copy()
        x_[~S] = self.reference_repeat[~S]

        return x_


class MarginalImputer:
    '''
    Impute features using either a fixed set of background examples, or a
    sampled set of background examples.

    Args:
      data: np.ndarray of data points representing the background distribution.
      samples: number of samples to draw (if None, then all samples are used.)
    '''
    def __init__(self, data, samples=None):
        self.data = data
        self.N = len(data)

        if samples is None:
            samples = self.N
        elif not samples < self.N:
            raise ValueError('The specified number of samples ({}) for '
                             'MarginalImputer should be less than the length '
                             'of the dataset ({}). Either use a smaller number '
                             'of samples, or set samples=None to use all '
                             'examples'.format(samples, self.N))
        self.samples = samples
        self.num_groups = data.shape[1]

        # For saving time during imputation.
        self.x_addr = None
        self.x_repeat = None
        if samples == self.N:
            self.data_tiled = data

        # Check if there are too many samples.
        if samples > 512:
            warnings.warn('using {} background samples will make estimators '
                          'run slowly, recommendation is to use <= 512'.format(
                            samples), RuntimeWarning)

    def __call__(self, x, S):
        # Repeat x.
        n = len(x)
        if self.x_addr == id(x):
            x = self.x_repeat
        else:
            self.x_addr = id(x)
            x = x.repeat(self.samples, 0)
            self.x_repeat = x

        # Prepare background samples.
        if self.samples == self.N:
            # Repeat fixed background samples.
            if len(self.data_tiled) != (self.samples * n):
                self.data_tiled = np.tile(self.data, (n, 1))
            samples = self.data_tiled
        else:
            # Draw samples from the marginal distribution.
            samples = self.data[np.random.choice(self.N, len(x), replace=True)]

        # Replace specified features.
        S = S.repeat(self.samples, 0)
        x_ = x.copy()
        x_[~S] = samples[~S]
        return x_



import pickle
import numpy as np
from sage import plotting


class SAGE:
    '''For storing and plotting SAGE values.'''
    def __init__(self, values, std):
        self.values = values
        self.std = std

    def plot(self,
             feature_names=None,
             sort_features=True,
             max_features=np.inf,
             orientation='horizontal',
             error_bars=True,
             color='tab:green',
             title='Feature Importance',
             title_size=20,
             tick_size=16,
             tick_rotation=None,
             label_size=16,
             figsize=(10, 7),
             return_fig=False):
        '''
        Plot SAGE values.

        Args:
          feature_names: list of feature names.
          sort_features: whether to sort features by their SAGE values.
          max_features: number of features to display.
          orientation: horizontal (default) or vertical.
          error_bars: whether to include standard deviation error bars.
          color: bar chart color.
          title: plot title.
          title_size: font size for title.
          tick_size: font size for feature names and numerical values.
          tick_rotation: tick rotation for feature names (vertical plots only).
          label_size: font size for label.
          figsize: figure size (if fig is None).
          return_fig: whether to return matplotlib figure object.
        '''
        return plotting.plot(
            self, feature_names, sort_features, max_features, orientation,
            error_bars, color, title, title_size, tick_size, tick_rotation,
            label_size, figsize, return_fig)

    def comparison(self,
                   other_values,
                   comparison_names=None,
                   feature_names=None,
                   sort_features=True,
                   max_features=np.inf,
                   orientation='vertical',
                   error_bars=True,
                   colors=None,
                   title='Feature Importance Comparison',
                   title_size=20,
                   tick_size=16,
                   tick_rotation=None,
                   label_size=16,
                   legend_loc=None,
                   figsize=(10, 7),
                   return_fig=False):
        '''
        Plot comparison with another set of SAGE values.

        Args:
          other_values: another SAGE values object.
          comparison_names: tuple of names for each SAGE value object.
          feature_names: list of feature names.
          sort_features: whether to sort features by their SAGE values.
          max_features: number of features to display.
          orientation: horizontal (default) or vertical.
          error_bars: whether to include standard deviation error bars.
          colors: colors for each set of SAGE values.
          title: plot title.
          title_size: font size for title.
          tick_size: font size for feature names and numerical values.
          tick_rotation: tick rotation for feature names (vertical plots only).
          label_size: font size for label.
          legend_loc: legend location.
          figsize: figure size (if fig is None).
          return_fig: whether to return matplotlib figure object.
        '''
        return plotting.comparison_plot(
            (self, other_values), comparison_names, feature_names,
            sort_features, max_features, orientation, error_bars, colors, title,
            title_size, tick_size, tick_rotation, label_size, legend_loc,
            figsize, return_fig)

    def save(self, filename):
        '''Save SAGE object.'''
        if isinstance(filename, str):
            with open(filename, 'wb') as f:
                pickle.dump(self, f)
        else:
            raise TypeError('filename must be str')


def load(filename):
    '''Load SAGE object.'''
    with open(filename, 'rb') as f:
        sage_values = pickle.load(f)
        if isinstance(sage_values, SAGE):
            return sage_values
        else:
            raise ValueError('object is not instance of SAGE class')



import warnings
import numpy as np
import matplotlib.pyplot as plt


def plot(sage_values,
         feature_names=None,
         sort_features=True,
         max_features=np.inf,
         orientation='horizontal',
         error_bars=True,
         color='tab:green',
         title='Feature Importance',
         title_size=20,
         tick_size=16,
         tick_rotation=None,
         label_size=16,
         figsize=(10, 7),
         return_fig=False):
    '''
    Plot SAGE values.

    Args:
      sage_values: SAGE values object.
      feature_names: list of feature names.
      sort_features: whether to sort features by their SAGE values.
      max_features: number of features to display.
      orientation: horizontal (default) or vertical.
      error_bars: whether to include standard deviation error bars.
      color: bar chart color.
      title: plot title.
      title_size: font size for title.
      tick_size: font size for feature names and numerical values.
      tick_rotation: tick rotation for feature names (vertical plots only).
      label_size: font size for label.
      figsize: figure size (if fig is None).
      return_fig: whether to return matplotlib figure object.
    '''
    # Default feature names.
    if feature_names is None:
        feature_names = ['Feature {}'.format(i) for i in
                         range(len(sage_values.values))]

    # Sort features if necessary.
    if len(feature_names) > max_features:
        sort_features = True

    # Perform sorting.
    values = sage_values.values
    std = sage_values.std
    if sort_features:
        argsort = np.argsort(values)[::-1]
        values = values[argsort]
        std = std[argsort]
        feature_names = np.array(feature_names)[argsort]

    # Remove extra features if necessary.
    if len(feature_names) > max_features:
        feature_names = (list(feature_names[:max_features])
                         + ['Remaining Features'])
        values = (list(values[:max_features])
                  + [np.sum(values[max_features:])])
        std = (list(std[:max_features])
               + [np.sum(std[max_features:] ** 2) ** 0.5])

    # Warn if too many features.
    if len(feature_names) > 50:
        warnings.warn('Plotting {} features may make figure too crowded, '
                      'consider using max_features'.format(
                        len(feature_names)), Warning)

    # Discard std if necessary.
    if not error_bars:
        std = None

    # Make plot.
    fig = plt.figure(figsize=figsize)
    ax = fig.gca()

    if orientation == 'horizontal':
        # Bar chart.
        ax.barh(np.arange(len(feature_names))[::-1], values,
                color=color, xerr=std)

        # Feature labels.
        if tick_rotation is not None:
            raise ValueError('rotation not supported for horizontal charts')
        ax.set_yticks(np.arange(len(feature_names))[::-1])
        ax.set_yticklabels(feature_names, fontsize=label_size)

        # Axis labels and ticks.
        ax.set_ylabel('')
        ax.set_xlabel('SAGE value', fontsize=label_size)
        ax.tick_params(axis='x', labelsize=tick_size)

    elif orientation == 'vertical':
        # Bar chart.
        ax.bar(np.arange(len(feature_names)), values, color=color,
               yerr=std)

        # Feature labels.
        if tick_rotation is None:
            tick_rotation = 45
        if tick_rotation < 90:
            ha = 'right'
            rotation_mode = 'anchor'
        else:
            ha = 'center'
            rotation_mode = 'default'
        ax.set_xticks(np.arange(len(feature_names)))
        ax.set_xticklabels(feature_names, rotation=tick_rotation, ha=ha,
                           rotation_mode=rotation_mode,
                           fontsize=label_size)

        # Axis labels and ticks.
        ax.set_ylabel('SAGE value', fontsize=label_size)
        ax.set_xlabel('')
        ax.tick_params(axis='y', labelsize=tick_size)

    else:
        raise ValueError('orientation must be horizontal or vertical')

    # Remove spines.
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    ax.set_title(title, fontsize=title_size)
    plt.tight_layout()

    if return_fig:
        return fig
    else:
        return


def comparison_plot(comparison_values,
                    comparison_names=None,
                    feature_names=None,
                    sort_features=True,
                    max_features=np.inf,
                    orientation='vertical',
                    error_bars=True,
                    colors=('tab:green', 'tab:blue'),
                    title='Feature Importance Comparison',
                    title_size=20,
                    tick_size=16,
                    tick_rotation=None,
                    label_size=16,
                    legend_loc=None,
                    figsize=(10, 7),
                    return_fig=False):
    '''
    Plot comparison between two different SAGE value objects.

    Args:
      comparison_values: tuple of two SAGE value objects to be compared.
      comparison_names: tuple of names for each SAGE value object.
      feature_names: list of feature names.
      sort_features: whether to sort features by their SAGE values.
      max_features: number of features to display.
      orientation: horizontal (default) or vertical.
      error_bars: whether to include standard deviation error bars.
      colors: colors for each set of SAGE values.
      title: plot title.
      title_size: font size for title.
      tick_size: font size for feature names and numerical values.
      tick_rotation: tick rotation for feature names (vertical plots only).
      label_size: font size for label.
      legend_loc: legend location.
      figsize: figure size (if fig is None).
      return_fig: whether to return matplotlib figure object.
    '''
    # Default feature names.
    if feature_names is None:
        feature_names = ['Feature {}'.format(i) for i in
                         range(len(comparison_values[0].values))]

    # Default comparison names.
    num_comps = len(comparison_values)
    if num_comps not in (2, 3, 4, 5):
        raise ValueError('only support comparisons for 2-5 sets of '
                         'SAGE values')
    if comparison_names is None:
        comparison_names = ['SAGE values {}'.format(i) for i in
                            range(len(num_comps))]

    # Default colors.
    if colors is None:
        colors = ['tab:green', 'tab:blue', 'tab:purple',
                  'tab:orange', 'tab:pink'][:num_comps]

    # Sort features if necessary.
    if len(feature_names) > max_features:
        sort_features = True

    # Extract values.
    values = [sage_values.values for sage_values in comparison_values]
    std = [sage_values.std for sage_values in comparison_values]

    # Perform sorting.
    if sort_features:
        argsort = np.argsort(values[0])[::-1]
        values = [sage_values[argsort] for sage_values in values]
        std = [stddev[argsort] for stddev in std]
        feature_names = np.array(feature_names)[argsort]

    # Remove extra features if necessary.
    if len(feature_names) > max_features:
        feature_names = (list(feature_names[:max_features])
                         + ['Remaining Features'])
        values = [
            list(sage_values[:max_features])
            + [np.sum(sage_values[max_features:])]
            for sage_values in values]
        std = [list(stddev[:max_features])
               + [np.sum(stddev[max_features:] ** 2) ** 0.5]
               for stddev in std]

    # Warn if too many features.
    if len(feature_names) > 50:
        warnings.warn('Plotting {} features may make figure too crowded, '
                      'consider using max_features'.format(
                        len(feature_names)), Warning)

    # Discard std if necessary.
    if not error_bars:
        std = [None for _ in std]

    # Make plot.
    width = 0.8 / num_comps
    fig = plt.figure(figsize=figsize)
    ax = fig.gca()

    if orientation == 'horizontal':
        # Bar chart.
        enumeration = enumerate(zip(values, std, comparison_names, colors))
        for i, (sage_values, stddev, name, color) in enumeration:
            pos = - 0.4 + width / 2 + width * i
            ax.barh(np.arange(len(feature_names))[::-1] - pos,
                    sage_values, height=width, color=color, xerr=stddev,
                    label=name)

        # Feature labels.
        if tick_rotation is not None:
            raise ValueError('rotation not supported for horizontal charts')
        ax.set_yticks(np.arange(len(feature_names))[::-1])
        ax.set_yticklabels(feature_names, fontsize=label_size)

        # Axis labels and ticks.
        ax.set_ylabel('')
        ax.set_xlabel('SAGE value', fontsize=label_size)
        ax.tick_params(axis='x', labelsize=tick_size)

    elif orientation == 'vertical':
        # Bar chart.
        enumeration = enumerate(zip(values, std, comparison_names, colors))
        for i, (sage_values, stddev, name, color) in enumeration:
            pos = - 0.4 + width / 2 + width * i
            ax.bar(np.arange(len(feature_names)) + pos,
                   sage_values, width=width, color=color, yerr=stddev,
                   label=name)

        # Feature labels.
        if tick_rotation is None:
            tick_rotation = 45
        if tick_rotation < 90:
            ha = 'right'
            rotation_mode = 'anchor'
        else:
            ha = 'center'
            rotation_mode = 'default'
        ax.set_xticks(np.arange(len(feature_names)))
        ax.set_xticklabels(feature_names, rotation=tick_rotation, ha=ha,
                           rotation_mode=rotation_mode,
                           fontsize=label_size)

        # Axis labels and ticks.
        ax.set_ylabel('SAGE value', fontsize=label_size)
        ax.set_xlabel('')
        ax.tick_params(axis='y', labelsize=tick_size)

    # Remove spines.
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    plt.legend(loc=legend_loc, fontsize=label_size)
    ax.set_title(title, fontsize=title_size)
    plt.tight_layout()

    if return_fig:
        return fig
    else:
        return



import numpy as np
import warnings
from copy import deepcopy


def verify_nonoverlapping(groups):
    '''Verify that no index is present in more than one group.'''
    used_inds = np.concatenate(groups)
    return np.all(np.unique(used_inds, return_counts=True)[1] == 1)


def finalize_groups(groups, input_size, remaining_features):
    '''Determine how to handle remaining features.'''
    assert remaining_features in ('split', 'group', 'ignore')
    remaining = np.setdiff1d(list(range(input_size)), np.concatenate(groups))
    if len(remaining) > 0:
        if remaining_features == 'split':
            for ind in remaining:
                groups.append([ind])
        elif remaining_features == 'group':
            groups.append(remaining_features)
    return groups


class GroupedReferenceImputer:
    '''
    Impute grouped features using reference values.

    Args:
      reference: the reference value for replacing missing features.
      groups: feature groups (list of lists).
      remaining_features: how to handle features that are not group members.
    '''
    def __init__(self, reference, groups, remaining_features='split'):
        if reference.ndim == 1:
            reference = reference[np.newaxis]
        elif reference.ndim == 2:
            if reference.shape[0] > 1:
                raise ValueError('expecting one set of reference values, '
                                 'got {}'.format(reference.shape[0]))
        self.reference = reference
        self.reference_repeat = self.reference
        self.samples = 1

        # Verify that groups are non-overlapping.
        input_size = reference.shape[1]
        groups = deepcopy(groups)
        nonoverlapping = verify_nonoverlapping(groups)
        if not nonoverlapping:
            raise ValueError('groups must be non-overlapping')

        # Handle remaining features.
        groups = finalize_groups(groups, input_size, remaining_features)
        self.groups = groups
        self.num_groups = len(groups)

        # Groups matrix.
        self.groups_mat = np.zeros((len(groups), input_size), dtype=bool)
        for i, group in enumerate(groups):
            self.groups_mat[i, group] = 1

    def __call__(self, x, S):
        # Repeat reference.
        if len(self.reference_repeat) != len(x):
            self.reference_repeat = self.reference.repeat(len(x), 0)

        # Convert groups to feature indices.
        S = np.matmul(S, self.groups_mat)

        x_ = x.copy()
        x_[~S] = self.reference.repeat(len(x), 0)[~S]
        return x_


class GroupedMarginalImputer:
    '''
    Impute grouped features using either a fixed set of background examples, or
    a sampled set of background examples.

    Args:
      data: np.ndarray of size (samples, dimensions) representing the data
        distribution.
      groups: feature groups (list of lists).
      remaining_features: how to handle features that are not group members.
    '''
    def __init__(self, data, groups, samples=None, remaining_features='split'):
        self.data = data
        self.N = len(data)

        if samples is None:
            samples = self.N
        elif not samples < self.N:
            raise ValueError('The specified number of samples ({}) for '
                             'GroupedMarginalImputer should be less than the '
                             'length of the dataset ({}). Either use a smaller '
                             'number of samples, or set samples=None to use '
                             'all examples'.format(samples, self.N))
        self.samples = samples

        # For saving time during imputation.
        self.x_addr = None
        self.x_repeat = None
        if samples == self.N:
            self.data_tiled = data

        # Check if there are too many samples.
        if samples > 512:
            warnings.warn('using {} background samples will make estimators '
                          'run slowly, recommendation is to use <= 512'.format(
                            samples), RuntimeWarning)

        # Verify that groups are non-overlapping.
        input_size = data.shape[1]
        groups = deepcopy(groups)
        nonoverlapping = verify_nonoverlapping(groups)
        if not nonoverlapping:
            raise ValueError('groups must be non-overlapping')

        # Handle remaining features.
        groups = finalize_groups(groups, input_size, remaining_features)
        self.groups = groups
        self.num_groups = len(groups)

        # Groups matrix.
        self.groups_mat = np.zeros((len(groups), input_size), dtype=bool)
        for i, group in enumerate(groups):
            self.groups_mat[i, group] = 1

    def __call__(self, x, S):
        # Repeat x.
        n = len(x)
        if self.x_addr == id(x):
            x = self.x_repeat
        else:
            self.x_addr = id(x)
            x = x.repeat(self.samples, 0)
            self.x_repeat = x

        # Prepare background samples.
        if self.N == self.samples:
            # Repeat data.
            if len(self.data_tiled) != (self.samples * n):
                self.data_tiled = np.tile(self.data, (n, 1))
            samples = self.data_tiled
        else:
            # Draw samples from marginal distribution.
            samples = self.data[np.random.choice(self.N, len(x), replace=True)]

        # Convert groups to feature indices.
        S = np.matmul(S, self.groups_mat)

        # Replace specified features.
        S = S.repeat(self.samples, 0)
        x_ = x.copy()
        x_[~S] = samples[~S]
        return x_



from sage import utils, core, imputers, grouped_imputers, plotting
from .core import SAGE, load
from .plotting import plot, comparison_plot
from .imputers import ReferenceImputer, MarginalImputer
from .grouped_imputers import GroupedReferenceImputer, GroupedMarginalImputer
from .permutation_sampler import PermutationSampler
from .iterated_sampler import IteratedSampler



import numpy as np
from sage import utils, core
from tqdm.auto import tqdm


def estimate_total(model, X, Y, batch_size, loss_fn):
    '''
    Estimate sum of SAGE values.

    This is used to get a worst-case estimate of the maximum
    SAGE value. It assumes that the prediction given no features is
    the mean prediction. This is not true of all imputers, and it will
    overestimate the total value in situations where certain features are
    never held out (which may lead to premature convergence).
    '''
    # Mean loss.
    N = 0
    mean_loss = 0
    mean_pred = 0
    for i in range(np.ceil(len(X) / batch_size).astype(int)):
        x = X[i * batch_size:(i + 1) * batch_size]
        y = Y[i * batch_size:(i + 1) * batch_size]
        n = len(x)
        pred = model(x)
        _mean_loss = np.mean(loss_fn(pred, y))
        _mean_pred = np.mean(pred, axis=0, keepdims=True)
        mean_loss = (N * mean_loss + n * _mean_loss) / (N + n)
        mean_pred = (N * mean_pred + n * _mean_pred) / (N + n)
        N += n

    # Mean loss with no features.
    N = 0
    marginal_loss = 0
    for i in range(np.ceil(len(X) / batch_size).astype(int)):
        y = Y[i * batch_size:(i + 1) * batch_size]
        n = len(x)
        # pred = mean_pred.repeat(n, *[1 for _ in mean_pred.shape[1:]])
        pred = mean_pred.repeat(n, 0)
        _mean_loss = np.mean(loss_fn(pred, y))
        marginal_loss = (N * marginal_loss + n * _mean_loss) / (N + n)
        N += n

    return marginal_loss - mean_loss


def estimate_holdout_importance(model, X, Y, imputer, batch_size, loss_fn,
                                batches, bar):
    '''
    Estimate impact of holding out features individually.

    This provides a rough estimate, but the result is only used to
    determine feature ordering.
    '''
    N, _ = X.shape
    num_features = imputer.num_groups
    holdout_importance = np.zeros(num_features)
    S = np.ones((batch_size, num_features), dtype=bool)

    if bar:
        bar = tqdm(total=num_features)

    # Hold out each feature individually.
    for i in range(num_features):
        S[:, i] = 0
        loss_list = []

        # Multiple minibatches to reduce variance.
        for _ in range(batches):
            mb = np.random.choice(N, batch_size)
            x = X[mb]
            y = Y[mb]
            pred = model(imputer(x, S))
            pred = np.mean(pred.reshape(
                -1, imputer.samples, *pred.shape[1:]), axis=1)
            loss = loss_fn(pred, y)
            loss_list.append(np.mean(loss))

        # Omit constant term (loss with all features).
        holdout_importance[i] = np.mean(loss_list)
        S[:, i] = 1
        if bar:
            bar.update(1)

    if bar:
        bar.close()
    return holdout_importance


class IteratedSampler:
    '''
    Estimate SAGE values one at a time by sampling subsets of features.

    Args:
      model: callable prediction model.
      imputer: for imputing held out values.
      loss: loss function ('mse', 'cross entropy').
    '''
    def __init__(self,
                 model,
                 imputer,
                 loss='cross entropy'):
        self.imputer = imputer
        self.loss_fn = utils.get_loss(loss, reduction='none')
        self.model = utils.model_conversion(model, self.loss_fn)

    def __call__(self,
                 X,
                 Y,
                 batch_size=512,
                 detect_convergence=True,
                 convergence_threshold=0.05,
                 n_samples=None,
                 optimize_ordering=True,
                 ordering_batches=1,
                 verbose=True,
                 bar=True):
        '''
        Estimate SAGE values.

        Args:
          X: input data.
          Y: target data.
          batch_size: number of examples to be processed in parallel, should be
            set to a large value.
          detect_convergence: whether to stop when approximately converged.
          convergence_threshold: threshold for determining convergence,
            represents ratio of standard deviation to max SAGE value.
          n_samples: number of samples to take per feature.
          optimize_ordering: whether to guess an ordering of features based on
            importance. May accelerate convergence.
          ordering_batches: number of minibatches while determining ordering.
          verbose: print progress messages.
          bar: display progress bar.

        The default behavior is to detect each feature's convergence based on
        the ratio of its standard deviation to the maximum value. Since the
        maximum value is unknown, we begin with a worst-case estimate
        (min_max_val) and update this estimate as more features are examined.

        Returns: SAGE object.
        '''
        # Verify model.
        N, _ = X.shape
        num_features = self.imputer.num_groups
        X, Y = utils.verify_model_data(self.model, X, Y, self.loss_fn,
                                       batch_size * self.imputer.samples)

        # For setting up bar.
        estimate_convergence = n_samples is None
        if estimate_convergence and verbose:
            print('Estimating convergence time')

        # Possibly force convergence detection.
        if n_samples is None:
            n_samples = 1e20
            if not detect_convergence:
                detect_convergence = True
                if verbose:
                    print('Turning convergence detection on')

        if detect_convergence:
            assert 0 < convergence_threshold < 1

        # Print message explaining parameter choices.
        if verbose:
            print('Batch size = batch * samples = {}'.format(
                batch_size * self.imputer.samples))

        # For detecting convergence.
        print('Estimating total importance')
        total = estimate_total(
            self.model, X, Y, batch_size * self.imputer.samples, self.loss_fn)
        min_max_val = total / num_features

        # Feature ordering.
        if optimize_ordering:
            if verbose:
                print('Determining feature ordering')
            holdout_importance = estimate_holdout_importance(
                self.model, X, Y, self.imputer, batch_size, self.loss_fn,
                ordering_batches, bar)
            ordering = list(np.argsort(holdout_importance)[::-1])
        else:
            ordering = list(range(num_features))

        # Set up bar.
        n_loops = int(n_samples / batch_size)
        if bar:
            if estimate_convergence:
                bar = tqdm(total=1)
            else:
                bar = tqdm(total=n_loops * batch_size * num_features)

        # Iterated sampling.
        tracker_list = []
        for i, ind in enumerate(ordering):
            tracker = utils.ImportanceTracker()
            for it in range(n_loops):
                # Sample data.
                mb = np.random.choice(N, batch_size)
                x = X[mb]
                y = Y[mb]

                # Sample subset of features.
                S = utils.sample_subset_feature(num_features, batch_size, ind)

                # Loss with feature excluded.
                y_hat = self.model(self.imputer(x, S))
                y_hat = np.mean(y_hat.reshape(
                    -1, self.imputer.samples, *y_hat.shape[1:]), axis=1)
                loss_discluded = self.loss_fn(y_hat, y)

                # Loss with feature included.
                S[:, ind] = 1.0
                y_hat = self.model(self.imputer(x, S))
                y_hat = np.mean(y_hat.reshape(
                    -1, self.imputer.samples, *y_hat.shape[1:]), axis=1)
                loss_included = self.loss_fn(y_hat, y)

                # Calculate delta sample.
                tracker.update(loss_discluded - loss_included)
                if bar and (not estimate_convergence):
                    bar.update(batch_size)

                # Calculate progress.
                std = tracker.std
                max_val = max(min_max_val, np.abs(tracker.values.item()))
                ratio = std / max_val

                # Print progress message.
                if verbose:
                    if detect_convergence:
                        print('StdDev Ratio = {:.4f} '
                              '(Converge at {:.4f})'.format(
                               ratio, convergence_threshold))
                    else:
                        print('StdDev Ratio = {:.4f}'.format(ratio))

                # Check for convergence.
                if detect_convergence:
                    if ratio < convergence_threshold:
                        if verbose:
                            print('Detected feature convergence')

                        # Skip bar ahead.
                        if bar:
                            bar.n = bar.total * (i + 1) / num_features
                            bar.refresh()
                        break

                # Update convergence estimation.
                if bar and estimate_convergence:
                    std_est = ratio * np.sqrt(it + 1)
                    n_est = (std_est / convergence_threshold) ** 2
                    bar.n = np.around((i + (it + 1) / n_est) / num_features, 4)
                    bar.refresh()

            if verbose:
                print('Done with feature {}'.format(i))
            tracker_list.append(tracker)

            # Adjust min max value.
            min_max_val = max(min_max_val, np.abs(tracker.values.item()))

        if bar:
            bar.close()

        # Extract SAGE values.
        reverse_ordering = [ordering.index(ind) for ind in range(num_features)]
        values = np.array(
            [tracker_list[ind].values.item() for ind in reverse_ordering])
        std = np.array(
            [tracker_list[ind].std.item() for ind in reverse_ordering])

        return core.SAGE(values, std)



import numpy as np
from sage import utils, core
from tqdm.auto import tqdm


class PermutationSampler:
    '''
    Estimate SAGE values by unrolling permutations of feature indices.

    Args:
      model: callable prediction model.
      imputer: for imputing held out values.
      loss: loss function ('mse', 'cross entropy').
    '''
    def __init__(self,
                 model,
                 imputer,
                 loss='cross entropy'):
        self.imputer = imputer
        self.loss_fn = utils.get_loss(loss, reduction='none')
        self.model = utils.model_conversion(model, self.loss_fn)

    def __call__(self,
                 X,
                 Y,
                 batch_size=512,
                 detect_convergence=True,
                 convergence_threshold=0.05,
                 n_permutations=None,
                 verbose=True,
                 bar=True):
        '''
        Estimate SAGE values.

        Args:
          X: input data.
          Y: target data.
          batch_size: number of examples to be processed in parallel, should be
            set to a large value.
          detect_convergence: whether to stop when approximately converged.
          convergence_threshold: threshold for determining convergence,
            represents ratio of max standard deviation to max SAGE value.
          n_permutations: number of permutations to unroll.
          verbose: print progress messages.
          bar: display progress bar.

        The default behavior is to detect convergence based on the width of the
        SAGE values' confidence intervals. Convergence is defined by the ratio
        of the maximum standard deviation to the maximum value reaching the
        specified threshold.

        Returns: SAGE object.
        '''
        # Verify model.
        N, _ = X.shape
        num_features = self.imputer.num_groups
        X, Y = utils.verify_model_data(self.model, X, Y, self.loss_fn,
                                       batch_size * self.imputer.samples)

        # For setting up bar.
        estimate_convergence = n_permutations is None
        if estimate_convergence and verbose:
            print('Estimating convergence time')

        # Possibly force convergence detection.
        if n_permutations is None:
            n_permutations = 1e20
            if not detect_convergence:
                detect_convergence = True
                if verbose:
                    print('Turning convergence detection on')

        if detect_convergence:
            assert 0 < convergence_threshold < 1

        # Print message explaining parameter choices.
        if verbose:
            print('Batch size = batch * samples = {}'.format(
                batch_size * self.imputer.samples))

        # Set up bar.
        n_loops = int(n_permutations / batch_size)
        if bar:
            if estimate_convergence:
                bar = tqdm(total=1)
            else:
                bar = tqdm(total=n_loops * batch_size * num_features)

        # Permutation sampling.
        tracker = utils.ImportanceTracker()
        for it in range(n_loops):
            # Sample data.
            mb = np.random.choice(N, batch_size)
            x = X[mb]
            y = Y[mb]

            # Sample permutations.
            S = np.zeros((batch_size, num_features), dtype=bool)
            permutations = np.tile(np.arange(num_features), (batch_size, 1))
            for i in range(batch_size):
                np.random.shuffle(permutations[i])

            # Make prediction with missing features.
            y_hat = self.model(self.imputer(x, S))
            y_hat = np.mean(y_hat.reshape(
                -1, self.imputer.samples, *y_hat.shape[1:]), axis=1)
            prev_loss = self.loss_fn(y_hat, y)

            # Setup.
            arange = np.arange(batch_size)
            scores = np.zeros((batch_size, num_features))

            for i in range(num_features):
                # Add next feature.
                inds = permutations[:, i]
                S[arange, inds] = 1

                # Make prediction with missing features.
                y_hat = self.model(self.imputer(x, S))
                y_hat = np.mean(y_hat.reshape(
                    -1, self.imputer.samples, *y_hat.shape[1:]), axis=1)
                loss = self.loss_fn(y_hat, y)

                # Calculate delta sample.
                scores[arange, inds] = prev_loss - loss
                prev_loss = loss
                if bar and (not estimate_convergence):
                    bar.update(batch_size)

            # Update tracker.
            tracker.update(scores)

            # Calculate progress.
            std = np.max(tracker.std)
            val = np.max(np.abs(tracker.values))
            ratio = std / val

            # Print progress message.
            if verbose:
                if detect_convergence:
                    print('StdDev Ratio = {:.4f} (Converge at {:.4f})'.format(
                        ratio, convergence_threshold))
                else:
                    print('StdDev Ratio = {:.4f}'.format(ratio))

            # Check for convergence.
            if detect_convergence:
                if ratio < convergence_threshold:
                    if verbose:
                        print('Detected convergence')

                    # Skip bar ahead.
                    if bar:
                        bar.n = bar.total
                        bar.refresh()
                    break

            # Update convergence estimation.
            if bar and estimate_convergence:
                std_est = ratio * np.sqrt(it + 1)
                n_est = (std_est / convergence_threshold) ** 2
                bar.n = np.around((it + 1) / n_est, 4)
                bar.refresh()

        if bar:
            bar.close()

        return core.SAGE(tracker.values, tracker.std)



import setuptools

setuptools.setup(
    name="sage-importance",
    version="0.0.3",
    author="Ian Covert",
    author_email="icovert@cs.washington.edu",
    description="For calculating global feature importance using Shapley values.",
    long_description="""
        SAGE (Shapley Additive Global importancE) is a game theoretic approach 
        for understanding black-box machine learning models. It summarizes each 
        feature's importance based on the predictive power it contributes, and 
        it accounts for complex interactions by using the Shapley value from 
        cooperative game theory. See the 
        [GitHub page](https://github.com/iancovert/sage/) for examples, and see 
        the [paper](https://arxiv.org/abs/2004.00668) for more details.
    """,
    long_description_content_type="text/markdown",
    url="https://github.com/iancovert/sage/",
    packages=['sage'],
    install_requires=[
        'numpy',
        'matplotlib',
        'tqdm'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
)



import sys
import numpy as np


class ImportanceTracker:
    '''For tracking feature importance using a dynamic average.'''
    def __init__(self):
        self.first_moment = 0
        self.second_moment = 0
        self.N = 0

    def update(self, scores):
        n = len(scores)
        first_moment = np.mean(scores, axis=0)
        second_moment = np.mean(scores ** 2, axis=0)
        self.first_moment = (
            (self.N * self.first_moment + n * first_moment) / (n + self.N))
        self.second_moment = (
            (self.N * self.second_moment + n * second_moment) / (n + self.N))
        self.N += n

    @property
    def values(self):
        return self.first_moment

    @property
    def var(self):
        return (self.second_moment - self.first_moment ** 2) / self.N

    @property
    def std(self):
        return self.var ** 0.5


class MSELoss:
    '''MSE loss that sums over non-batch dimensions.'''
    def __init__(self, reduction='mean'):
        assert reduction in ('none', 'mean')
        self.reduction = reduction

    def __call__(self, pred, target):
        # Add dimension if necessary.
        if target.shape[-1] == 1 and len(target.shape) - len(pred.shape) == 1:
            pred = np.expand_dims(pred, -1)
        loss = np.sum(
            np.reshape((pred - target) ** 2, (len(pred), -1)), axis=1)
        if self.reduction == 'mean':
            return np.mean(loss)
        else:
            return loss


class CrossEntropyLoss:
    '''Cross entropy loss that expects probabilities.'''
    def __init__(self, reduction='mean'):
        assert reduction in ('none', 'mean')
        self.reduction = reduction

    def __call__(self, pred, target, eps=1e-12):
        # Clip.
        pred = np.clip(pred, eps, 1 - eps)

        # Add a dimension if necessary.
        if pred.ndim == 1:
            pred = pred[:, np.newaxis]
        if pred.shape[1] == 1:
            pred = np.append(1 - pred, pred, axis=1)

        # Calculate loss.
        loss = - np.log(pred[np.arange(len(pred)), target])
        if self.reduction == 'mean':
            return np.mean(loss)
        else:
            return loss


def get_loss(loss, reduction='mean'):
    '''Get loss function by name.'''
    if loss == 'cross entropy':
        loss_fn = CrossEntropyLoss(reduction=reduction)
    elif loss == 'mse':
        loss_fn = MSELoss(reduction=reduction)
    else:
        raise ValueError('unsupported loss: {}'.format(loss))
    return loss_fn


def sample_subset_feature(input_size, n, ind):
    '''
    Sample a subset of features where a given feature index must not be
    included. This helper function is used for estimating Shapley values, so
    the subset is sampled by 1) sampling the number of features to be included
    from a uniform distribution, and 2) sampling the features to be included.
    '''
    S = np.zeros((n, input_size), dtype=bool)
    choices = list(range(input_size))
    del choices[ind]
    for row in S:
        inds = np.random.choice(
            choices, size=np.random.choice(input_size), replace=False)
        row[inds] = 1
    return S


def verify_model_data(model, X, Y, loss, mbsize):
    '''Ensure that model and data are set up properly.'''
    if isinstance(loss, CrossEntropyLoss):
        probs = model(X[:mbsize])
        Y = Y.astype(int)

        if (probs.ndim == 1) or (probs.shape[1] == 1):
            # Single probabilities.
            if Y.shape == (len(X),):
                pass
            elif Y.shape == (len(X), 1):
                Y = Y[:, 0]
            else:
                raise ValueError('labels shape is incorrect')

            valid_probs = np.all(np.logical_and(probs >= 0, probs <= 1))
            unique_labels = np.unique(Y)
            if np.array_equal(unique_labels, np.array([0, 1])):
                # This is the preferred labeling.
                pass
            elif np.array_equal(unique_labels, np.array([-1, 1])):
                # Set -1 to 0.
                Y = Y.copy()
                Y[Y == -1] = 0
            else:
                raise ValueError('labels for binary classification must be '
                                 '[0, 1] or [-1, 1]')

        elif probs.ndim == 2:
            # Multiclass output.
            assert Y.shape == (len(X),)
            ones = np.sum(probs, axis=1)
            valid_probs = np.allclose(ones, np.ones(ones.shape))

        else:
            raise ValueError('predictions array has too many dimensions')

        if not valid_probs:
            raise ValueError('outputs must be valid probabilities')

    return X, Y


def safe_isinstance(obj, class_str):
    '''Check isinstance without requiring imports.'''
    if not isinstance(class_str, str):
        return False
    module_name, class_name = class_str.rsplit('.', 1)
    if module_name not in sys.modules:
        return False
    module = sys.modules[module_name]
    class_type = getattr(module, class_name, None)
    if class_type is None:
        return False
    return isinstance(obj, class_type)


def model_conversion(model, loss_fn):
    '''Convert model to callable.'''
    if safe_isinstance(model, 'sklearn.base.ClassifierMixin'):
        return lambda x: model.predict_proba(x)

    elif safe_isinstance(model, 'sklearn.base.RegressorMixin'):
        return lambda x: model.predict(x)

    elif safe_isinstance(model, 'catboost.CatBoostClassifier'):
        return lambda x: model.predict_proba(x)

    elif safe_isinstance(model, 'catboost.CatBoostRegressor'):
        return lambda x: model.predict(x)

    elif safe_isinstance(model, 'lightgbm.basic.Booster'):
        return lambda x: model.predict(x)

    elif safe_isinstance(model, 'xgboost.core.Booster'):
        import xgboost
        return lambda x: model.predict(xgboost.DMatrix(x))

    elif safe_isinstance(model, 'torch.nn.Module'):
        import torch
        device = next(model.parameters()).device

        if isinstance(loss_fn, CrossEntropyLoss):
            print('Converting PyTorch classifier, outputs are assumed '
                  'to be logits')

            def f(x):
                x = torch.tensor(x, dtype=torch.float32, device=device)
                out = model(x)
                if out.dim() == 1:
                    out = out.sigmoid()
                elif out.dim() == 2 and out.shape[1] == 1:
                    out = out.sigmoid()
                elif out.dim() == 2:
                    out = out.softmax(dim=1)
                else:
                    raise ValueError('predictions are not valid shape')
                return out.cpu().data.numpy()
        else:
            def f(x):
                x = torch.tensor(x, dtype=torch.float32, device=device)
                out = model(x)
                return out.cpu().data.numpy()

        return f

    elif callable(model):
        # Assume model is compatible function or callable object.
        return model

    else:
        raise ValueError('model cannot be converted automatically, '
                         'sorry! Please convert to a lambda function')
